/**
 * @file This file is for helper functions related to Pipcook-core
 */
import * as path from 'path';
import * as fs from 'fs-extra';
import * as request from 'request-promise';
import * as url from 'url';
import {
  RunConfigI,
  PipelineDBParams,
  constants,
  generateId,
  PluginTypeI
} from '@pipcook/pipcook-core';

const { PLUGINS, PIPCOOK_LOGS } = constants;

/**
 * The pipeline database schema.
 */
export declare type PipelineDB = Partial<Record<'id' | 'name' |
  'dataCollectId' | 'dataAccessId' | 'dataProcessId' | 'modelDefineId'
  | 'modelLoadId' | 'modelTrainId' | 'modelEvaluateId' | PluginTypeI | PipelineDBParams, string>>;

async function loadConfig(configPath: string | RunConfigI): Promise<RunConfigI> {
  if (typeof configPath === 'string') {
    let configJson: RunConfigI;
    const urlObj = url.parse(configPath);
    if (urlObj.protocol === null) {
      throw new TypeError('config URI is not supported');
    }
    if ([ 'http:', 'https:' ].indexOf(urlObj.protocol) >= 0) {
      configJson = JSON.parse(await request(configPath));
      for (const key in configJson.plugins) {
        const plugin = configJson.plugins[key];
        if (path.isAbsolute(plugin.package) || plugin.package.startsWith('.')) {
          throw new TypeError(`local path is invalid for plugin package: ${plugin.package}`);
        }
      }
    } else if (urlObj.protocol === 'file:') {
      configJson = await fs.readJSON(url.fileURLToPath(configPath));
    } else {
      throw new TypeError(`protocol ${urlObj.protocol} is not supported`);
    }
    return configJson;
  } else {
    return configPath;
  }
}

export async function parseConfig(configPath: string | RunConfigI, isGenerateId = true): Promise<PipelineDB> {
  const configJson = await loadConfig(configPath);
  const result: PipelineDB = {};

  if (isGenerateId) {
    result.id = generateId();
  }
  if (configJson.name) {
    result.name = configJson.name;
  }

  PLUGINS.forEach((pluginType) => {
    if (configJson.plugins[pluginType]?.package) {
      const pluginName = configJson.plugins[pluginType].package;
      const params = configJson.plugins[pluginType].params || {};

      result[pluginType] = pluginName;
      const paramsAttribute: PipelineDBParams = (pluginType + 'Params') as PipelineDBParams;
      result[paramsAttribute] = JSON.stringify(params);
    }
  });
  return result;
}

export async function writeOutput(jobId: string, content: string, stderr = false): Promise<void> {
  const filename = stderr ? 'stderr' : 'stdout';
  const dest = path.join(PIPCOOK_LOGS, jobId, filename);
  await fs.appendFile(dest, content);
}
